import 'dart:async';

import 'package:bloc_stream/bloc_stream.dart';
import 'package:riverpod/riverpod.dart';
import 'package:meta/meta.dart';

// ignore: implementation_imports
import 'package:riverpod/src/framework.dart';

import 'builders.dart';

mixin _BlocStreamStateProviderStateMixin<T>
    on ProviderStateBase<BlocStream<T>, T> {
  StreamSubscription<T>? subscription;

  @override
  void valueChanged({BlocStream<T>? previous}) {
    if (createdValue == previous) {
      return;
    }
    subscription?.cancel();
    _listener(createdValue.value);
    subscription = createdValue.skip(1).listen(_listener);
  }

  void _listener(T value) {
    exposedValue = value;
  }

  @override
  void dispose() {
    subscription?.cancel();
    super.dispose();
  }
}

// ignore: subtype_of_sealed_class
@sealed
class BlocStreamProvider<T extends BlocStream<Object?>> extends Provider<T> {
  BlocStreamProvider(
    Create<T, ProviderReference> create, {
    String? name,
  }) : super((ref) {
          final bloc = create(ref);
          ref.onDispose(bloc.close);
          return bloc;
        }, name: name);

  static const family = BlocStreamProviderFamilyBuilder();
  static const autoDispose = AutoDisposeBlocStreamProviderBuilder();

  BlocStreamStateProvider<Object?>? _value;
}

/// Adds [value] to [BlocStreamProvider.autoDispose].
extension BlocStreamStateProviderX<Value>
    on BlocStreamProvider<BlocStream<Value>> {
  BlocStreamStateProvider<Value> get value {
    _value ??= BlocStreamStateProvider<Value>._(this);
    return _value as BlocStreamStateProvider<Value>;
  }
}

@sealed
class BlocStreamStateProvider<T>
    extends AlwaysAliveProviderBase<BlocStream<T>, T> {
  BlocStreamStateProvider._(this._provider)
      : super(
          (ref) => ref.watch(_provider),
          _provider.name != null ? '${_provider.name}.value' : null,
        );

  final BlocStreamProvider<BlocStream<T>> _provider;

  @override
  _BlocStreamStateProviderState<T> createState() {
    return _BlocStreamStateProviderState<T>();
  }

  @override
  Override overrideWithValue(T value) {
    return ProviderOverride(
      ValueProvider<BlocStream<T>, T>((ref) {
        return ref.watch(_provider);
      }, value),
      this,
    );
  }
}

@sealed
class _BlocStreamStateProviderState<T> = ProviderStateBase<BlocStream<T>, T>
    with _BlocStreamStateProviderStateMixin<T>;

@sealed
class BlocStreamProviderFamily<T extends BlocStream<Object?>, A>
    extends Family<T, T, A, ProviderReference, BlocStreamProvider<T>> {
  BlocStreamProviderFamily(
    T Function(ProviderReference ref, A a) create, {
    String? name,
  }) : super(create, name);

  @override
  BlocStreamProvider<T> create(
    A value,
    T Function(ProviderReference ref, A param) builder,
    String? name,
  ) {
    return BlocStreamProvider((ref) => builder(ref, value), name: name);
  }
}

// ignore: subtype_of_sealed_class
@sealed
class AutoDisposeBlocStreamProvider<T extends BlocStream<Object?>>
    extends AutoDisposeProvider<T> {
  AutoDisposeBlocStreamProvider(
    Create<T, AutoDisposeProviderReference> create, {
    String? name,
  }) : super((ref) {
          final controller = create(ref);
          ref.onDispose(controller.close);
          return controller;
        }, name: name);

  static const family = AutoDisposeBlocStreamProviderFamilyBuilder();

  AutoDisposeBlocStreamStateProvider<Object?>? _value;
}

/// Adds [value] to [BlocStreamProvider.autoDispose].
extension AutoDisposeBlocStreamStateProviderX<Value>
    on AutoDisposeBlocStreamProvider<BlocStream<Value>> {
  AutoDisposeBlocStreamStateProvider<Value> get value {
    _value ??= AutoDisposeBlocStreamStateProvider<Value>._(this);
    return _value as AutoDisposeBlocStreamStateProvider<Value>;
  }
}

@sealed
class AutoDisposeBlocStreamStateProvider<T>
    extends AutoDisposeProviderBase<BlocStream<T>, T> {
  AutoDisposeBlocStreamStateProvider._(
      AutoDisposeBlocStreamProvider<BlocStream<T>> provider)
      : super(
          (ref) => ref.watch(provider),
          provider.name != null ? '${provider.name}.value' : null,
        );

  @override
  _AutoDisposeBlocStreamStateProviderState<T> createState() {
    return _AutoDisposeBlocStreamStateProviderState<T>();
  }
}

@sealed
class _AutoDisposeBlocStreamStateProviderState<T> = ProviderStateBase<
    BlocStream<T>, T> with _BlocStreamStateProviderStateMixin<T>;

@sealed
class AutoDisposeBlocStreamProviderFamily<T extends BlocStream<Object?>, A>
    extends Family<T, T, A, AutoDisposeProviderReference,
        AutoDisposeBlocStreamProvider<T>> {
  AutoDisposeBlocStreamProviderFamily(
    T Function(AutoDisposeProviderReference ref, A a) create, {
    String? name,
  }) : super(create, name);

  @override
  AutoDisposeBlocStreamProvider<T> create(
    A value,
    T Function(AutoDisposeProviderReference ref, A param) builder,
    String? name,
  ) {
    return AutoDisposeBlocStreamProvider(
      (ref) => builder(ref, value),
      name: name,
    );
  }
}
